// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';
import 'dart:isolate';

import 'package:wyatt_cli_toolbox/src/sequences/sequences.dart';
import 'package:wyatt_cli_toolbox/src/workflow/skip_exception.dart';
import 'package:wyatt_cli_toolbox/src/workflow/task.dart';

export 'skip_exception.dart';
export 'task.dart';
export 'utils.dart';

enum WorkflowControl { start, next, stop }

class Workflow {
  Workflow._(this._workflowIsolate, this._tasks, this._toIsolate);

  Isolate? _workflowIsolate;
  final SendPort _toIsolate;
  final List<Task<dynamic>> _tasks;
  final Map<String, dynamic> _results = <String, dynamic>{};

  Map<String, dynamic> get results => _results;

  static Future<Workflow> _initIsolate(List<Task<dynamic>> tasks) async {
    final List<IsolateTask> animations = <IsolateTask>[];
    for (final Task<dynamic> t in tasks) {
      animations.add(t.animation);
    }

    Isolate? workflowIsolate;

    final Completer<Workflow> completer = Completer<Workflow>();
    final ReceivePort isolateToMainStream = ReceivePort()
      // Retrieve send port sent by isolate
      ..listen((dynamic data) {
        if (data is SendPort) {
          completer.complete(Workflow._(workflowIsolate, tasks, data));
        }
      });

    workflowIsolate = await Isolate.spawn(
      _start,
      <Object>[isolateToMainStream.sendPort, animations],
    );
    return completer.future;
  }

  static void _start(List<Object?> args) {
    // ignore: cast_nullable_to_non_nullable
    final SendPort isolateToMainStream = args[0] as SendPort;
    // ignore: cast_nullable_to_non_nullable
    final List<IsolateTask> tasks = args[1] as List<IsolateTask>;
    final ReceivePort mainToIsolateStream = ReceivePort();

    isolateToMainStream.send(mainToIsolateStream.sendPort);

    mainToIsolateStream.listen((dynamic data) {
      final List<Object?> args = data as List<Object?>;
      // ignore: cast_nullable_to_non_nullable
      final WorkflowControl control = args[0] as WorkflowControl;
      // ignore: cast_nullable_to_non_nullable
      final int taskId = args[1] as int;
      final ExitStatus? exit = args[2] as ExitStatus?;
      final String? reason = args[3] as String?;

      if (control == WorkflowControl.start) {
        tasks[taskId].run();
      } else if (control == WorkflowControl.next) {
        if (exit != null) {
          tasks[taskId - 1].stop(exit, reason: reason);
        }
        tasks[taskId].run();
      } else if (control == WorkflowControl.stop) {
        tasks[taskId].stop(exit!, reason: reason);
      }
    });
  }

  static Future<Workflow> create(List<Task<dynamic>> tasks) async {
    final Workflow w = await _initIsolate(tasks);
    return w;
  }

  void _animation(
    WorkflowControl control,
    int taskId, {
    ExitStatus? status,
    String? reason,
  }) {
    _toIsolate.send(<Object?>[control, taskId, status, reason]);
  }

  Map<String, dynamic> start() {
    Sequences.hideCursor();
    int _index = 0;

    for (final Task<dynamic> _ in _tasks) {
      _call(_index);
      _index++;
    }

    if (_workflowIsolate != null) {
      _workflowIsolate!.kill(priority: Isolate.immediate);
    }
    _workflowIsolate = null;
    Sequences.unhideCursor();
    return _results;
  }

  ExitStatus _call(int taskId) {
    ExitStatus status = ExitStatus.nil;
    String message = 'nil';
    final Task<dynamic> t = _tasks[taskId];

    if (t.task != null) {
      _animation(WorkflowControl.start, taskId);
      try {
        // Call task function
        final dynamic res = t.invoke();
        _results[t.identifier] = res;

        message = 'Finished';
        if (t.onFinish != null) {
          message = t.onFinish!(res);
        }
        status = ExitStatus.success;
      } catch (e) {
        if (e is SkipException) {
          message = e.toString();
          if (t.onSkip != null) {
            message = t.onSkip!(e);
          }

          status = ExitStatus.skip;
        } else {
          message = e.toString();
          if (t.onError != null) {
            message = t.onError!(e);
          }

          status = ExitStatus.fail;
        }
      }
      _animation(WorkflowControl.stop, taskId, status: status, reason: message);
    }
    return status;
  }
}
